var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _ReservationStore_instances, _ReservationStore_createReservation;
import { PeerMap } from '@libp2p/peer-collections';
import PQueue from 'p-queue';
import { DEFAULT_RESERVATION_CONCURRENCY, RELAY_TAG, RELAY_V2_HOP_CODEC } from '../constants.js';
import { logger } from '@libp2p/logger';
import { pbStream } from 'it-pb-stream';
import { HopMessage, Status } from '../pb/index.js';
import { getExpirationMilliseconds } from '../utils.js';
import { CustomEvent, EventEmitter } from '@libp2p/interfaces/events';
import { multiaddr } from '@multiformats/multiaddr';
const log = logger('libp2p:circuit-relay:transport:reservation-store');
// allow refreshing a relay reservation if it will expire in the next 10 minutes
const REFRESH_WINDOW = (60 * 1000) * 10;
// try to refresh relay reservations 5 minutes before expiry
const REFRESH_TIMEOUT = (60 * 1000) * 5;
// minimum duration before which a reservation must not be refreshed
const REFRESH_TIMEOUT_MIN = 30 * 1000;
export class ReservationStore extends EventEmitter {
    constructor(components, init) {
        super();
        _ReservationStore_instances.add(this);
        this.peerId = components.peerId;
        this.connectionManager = components.connectionManager;
        this.transportManager = components.transportManager;
        this.peerStore = components.peerStore;
        this.reservations = new PeerMap();
        this.maxDiscoveredRelays = init?.discoverRelays ?? 0;
        this.started = false;
        // ensure we don't listen on multiple relays simultaneously
        this.reserveQueue = new PQueue({
            concurrency: init?.reservationConcurrency ?? DEFAULT_RESERVATION_CONCURRENCY
        });
        // When a peer disconnects, if we had a reservation on that peer
        // remove the reservation and multiaddr and maybe trigger search
        // for new relays
        this.connectionManager.addEventListener('peer:disconnect', (evt) => {
            this.removeRelay(evt.detail.remotePeer);
        });
    }
    isStarted() {
        return this.started;
    }
    async start() {
        this.started = true;
    }
    async stop() {
        this.reservations.forEach(({ timeout }) => {
            clearTimeout(timeout);
        });
        this.reservations.clear();
        this.started = true;
    }
    /**
     * If the number of current relays is beneath the configured `maxReservations`
     * value, and the passed peer id is not our own, and we have a non-relayed connection
     * to the remote, and the remote peer speaks the hop protocol, try to reserve a slot
     * on the remote peer
     */
    async addRelay(peerId, type) {
        log('add relay', this.reserveQueue.size);
        await this.reserveQueue.add(async () => {
            try {
                if (this.peerId.equals(peerId)) {
                    log('not trying to use self as relay');
                    return;
                }
                // allow refresh of an existing reservation if it is about to expire
                const existingReservation = this.reservations.get(peerId);
                if (existingReservation != null) {
                    if (getExpirationMilliseconds(existingReservation.reservation.expire) > REFRESH_WINDOW) {
                        log('already have reservation on relay peer %p and it expires in more than 10 minutes', peerId);
                        return;
                    }
                    clearTimeout(existingReservation.timeout);
                    this.reservations.delete(peerId);
                }
                if (type === 'discovered' && [...this.reservations.values()].reduce((acc, curr) => {
                    if (curr.type === 'discovered') {
                        acc++;
                    }
                    return acc;
                }, 0) >= this.maxDiscoveredRelays) {
                    log('already have enough discovered relays');
                    return;
                }
                const connection = await this.connectionManager.openConnection(peerId);
                if (connection.remoteAddr.protoNames().includes('p2p-circuit')) {
                    log('not creating reservation over relayed connection');
                    return;
                }
                const reservation = await __classPrivateFieldGet(this, _ReservationStore_instances, "m", _ReservationStore_createReservation).call(this, connection);
                log('created reservation on relay peer %p', peerId);
                const expiration = getExpirationMilliseconds(reservation.expire);
                // sets a lower bound on the timeout, and also don't let it go over
                // 2^31 - 1 (setTimeout will only accept signed 32 bit integers)
                const timeoutDuration = Math.min(Math.max(expiration - REFRESH_TIMEOUT, REFRESH_TIMEOUT_MIN), Math.pow(2, 31) - 1);
                const timeout = setTimeout(() => {
                    this.addRelay(peerId, type).catch(err => {
                        log.error('could not refresh reservation to relay %p', peerId, err);
                    });
                }, timeoutDuration);
                this.reservations.set(peerId, {
                    timeout,
                    reservation,
                    type
                });
                // ensure we don't close the connection to the relay
                await this.peerStore.tagPeer(peerId, RELAY_TAG, {
                    value: 1,
                    ttl: expiration
                });
                await this.transportManager.listen(reservation.addrs.map(ma => {
                    return multiaddr(ma).encapsulate('/p2p-circuit');
                }));
            }
            catch (err) {
                log.error('could not reserve slot on %p', peerId, err);
            }
        });
    }
    hasReservation(peerId) {
        return this.reservations.has(peerId);
    }
    /**
     * Remove listen relay
     */
    removeRelay(peerId) {
        const existingReservation = this.reservations.get(peerId);
        if (existingReservation == null) {
            return;
        }
        log('removing relay %p', peerId);
        clearTimeout(existingReservation.timeout);
        this.reservations.delete(peerId);
        this.safeDispatchEvent('relay:removed', { detail: peerId });
        if (this.reservations.size < this.maxDiscoveredRelays) {
            this.dispatchEvent(new CustomEvent('relay:not-enough-relays'));
        }
    }
}
_ReservationStore_instances = new WeakSet(), _ReservationStore_createReservation = async function _ReservationStore_createReservation(connection) {
    log('requesting reservation from %s', connection.remotePeer);
    const stream = await connection.newStream(RELAY_V2_HOP_CODEC);
    const pbstr = pbStream(stream);
    const hopstr = pbstr.pb(HopMessage);
    hopstr.write({ type: HopMessage.Type.RESERVE });
    let response;
    try {
        response = await hopstr.read();
    }
    catch (err) {
        log.error('error parsing reserve message response from %p because', connection.remotePeer, err);
        throw err;
    }
    finally {
        stream.close();
    }
    if (response.status === Status.OK && (response.reservation != null)) {
        return response.reservation;
    }
    const errMsg = `reservation failed with status ${response.status ?? 'undefined'}`;
    log.error(errMsg);
    throw new Error(errMsg);
};
//# sourceMappingURL=reservation-store.js.map